/* ContainerSelector.m created by benhur on 12-feb-2001 */

#include "ContainerSelector.h"
#include "../General.bproj/FoundationAdditions.h"
#include "../General.bproj/UniqueString.h"
#include "../General.bproj/MultiEnumerator.h"

@implementation ContainerSelector

- (id)initWithController:(PajeTraceController *)c
{
    NSWindow *win;
    
    [super initWithController:c];

    filters = [[NSMutableDictionary alloc] init];

    if (![NSBundle loadNibNamed:@"ContainerSelector" owner:self]) {
        NSRunAlertPanel(@"ContainerSelector", @"Couldn't load interface file",
                        nil, nil, nil);
        return self;
    }

    win = [view window];
    
    // view is an NSBox. we need its contents.
    view = [[view contentView] retain];
    [view removeFromSuperview];

    hierarchyBrowser = [[HierarchyBrowser alloc] initWithFilter:self];
    [hierarchyBrowser setContainersOnly:YES];
    [[hierarchyBrowser view] setFrame:[browserBox frame]];
    [[browserBox superview] replaceSubview:browserBox with:[hierarchyBrowser view]];

    [self registerFilter:self];

#ifndef GNUSTEP
    [win release];
#endif

    return self;
}

- (NSString *)filterName
{
    return @"Container Selector";
}

- (NSView *)filterView
{
    return view;
}

- (id)filterDelegate
{
    return self;
}

- (void)setInputComponent:(id)component
{
    [super setInputComponent:component];
    [hierarchyBrowser setFilter:component];
}

- (void)dealloc
{
    [filters release];
    [view release];
    [super dealloc];
}


//
// Notifications sent by other filters
//
- (void)hierarchyChanged
{
    PajeEntityType *entityType;
    NSEnumerator *typeEnum;
    
    typeEnum = [[inputComponent allEntityTypes] objectEnumerator];
    while ((entityType = [typeEnum nextObject]) != nil) {
        if ([self isContainerEntityType:entityType]) {
            [self setFilterForEntityType:entityType];
        }
    }
    [self synchronizeMatrix];
    [super hierarchyChanged];
}

- (void)colorChangedForEntityType:(PajeEntityType *)entityType
{
    if (!entityType || [entityType isEqual:[self selectedEntityType]])
        [self synchronizeMatrix];
    [super colorChangedForEntityType:entityType];
}



//
//
//
- (void)setFilterForEntityType:(PajeEntityType *)entityType
// a new entity type is known. create a new filter for it.
{
    NSMutableSet *filter;

    NSParameterAssert(entityType);
    
    filter = [filters objectForKey:entityType];

    if (filter == nil) {
        filter = [self defaultFilterForEntityType:entityType];
        [filters setObject:filter forKey:entityType];
    }
}

- (NSMutableSet *)defaultFilterForEntityType:(PajeEntityType *)entityType
{
    NSString *defaultName = [NSString stringWithFormat:@"%@%@Filter", NSStringFromClass([self class]), [entityType name]];
    NSArray *filterArray = [[NSUserDefaults standardUserDefaults] arrayForKey:defaultName];
    NSMutableSet *uniqueSet;
    NSEnumerator *filterEnum;
    NSString *s;
    
#ifdef GNUSTEP
    uniqueSet = [[[NSMutableSet alloc] init] autorelease];
#else
    uniqueSet = [NSMutableSet set];
#endif
    filterEnum = [filterArray objectEnumerator];
    while ((s = [filterEnum nextObject]) != nil) {
        [uniqueSet addObject:U(s)];
    }
    
    return uniqueSet;
}

- (void)registerDefaultFilter:(NSSet *)filter
                forEntityType:(PajeEntityType *)entityType
{
    NSString *defaultName = [NSString stringWithFormat:@"%@%@Filter", NSStringFromClass([self class]), [entityType name]];
    NSArray *filterArray = [filter allObjects];
    [[NSUserDefaults standardUserDefaults] setObject:filterArray
                                              forKey:defaultName];
    NSLog(@"registrou %@ = %@", defaultName, filterArray);
}

- (PajeEntityType *)selectedEntityType
{
    return [hierarchyBrowser selectedEntityType];
}


//
// Handle interface messages
//
- (void)selectAll:(id)sender
{
    PajeEntityType *entityType = [hierarchyBrowser selectedEntityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        [filter minusSet:[NSSet setWithArray:[hierarchyBrowser selectedNames]]];
//        [filter removeAllObjects];
        [self synchronizeMatrix];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [NSException raise:@"ContainerSelector: Internal Inconsistency"
                    format:@"%@ sent by unknown view %@", NSStringFromSelector(_cmd), sender];
    }
}

- (void)unselectAll:(id)sender
{
    PajeEntityType *entityType = [hierarchyBrowser selectedEntityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        [filter addObjectsFromArray:[hierarchyBrowser selectedNames]];
        [self synchronizeMatrix];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [NSException raise:@"ContainerSelector: Internal Inconsistency"
                    format:@"%@ sent by unknown view %@", NSStringFromSelector(_cmd), sender];
    }
}

- (void)matrixChanged:(id)sender
{
    NSButtonCell *cell = [sender selectedCell];
    id entityType = [hierarchyBrowser selectedEntityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (nil == cell)
        return;
    
    if (filter) {
        if ([cell state])
            [filter removeObject:[cell representedObject]];
        else
            [filter addObject:[cell representedObject]];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [NSException raise:@"ContainerSelector: Internal Inconsistency"
                    format:@"%@ sent by unknown view %@", NSStringFromSelector(_cmd), sender];
    }
}

- (void)entityTypeSelected:(id)sender
{
    [self synchronizeMatrix];
}

- (void)synchronizeMatrix
{
    [hierarchyBrowser setFilter:inputComponent];
}

//
// interact with interface
//

- (void)viewWillBeSelected:(NSView *)selectedView
// message sent by PajeController when a view will be selected for display
{
    [self synchronizeMatrix];        
}


//
// Trace messages that are filtered
//
- (NSEnumerator *)enumeratorOfContainersTyped:(PajeContainerType *)entityType
                                  inContainer:(PajeContainer *)container
{
    NSEnumerator *origEnum;
    NSMutableSet *filter;

    if ([[entityType containerType] isEqual:[container entityType]]
        || [entityType containerType] == (PajeContainerType *)entityType
        || [entityType containerType] == nil/*HACK*/) {
        origEnum = [super enumeratorOfContainersTyped:entityType
                                          inContainer:container];
    } else {
        NSEnumerator *subcontainerEnum;
        MultiEnumerator *multiEnum;
        id subcontainer;
	PajeContainerType *parentType = [entityType containerType];
        subcontainerEnum = [self enumeratorOfContainersTyped:parentType
                                                 inContainer:container];
        multiEnum = [MultiEnumerator enumerator];
        while ((subcontainer = [subcontainerEnum nextObject]) != nil) {
            NSEnumerator *enumerator;
            enumerator = [super enumeratorOfContainersTyped:entityType
                                                inContainer:subcontainer];
            if (enumerator != nil)
                [multiEnum addEnumerator:enumerator];
        }
        origEnum = multiEnum;
    }

    filter = [filters objectForKey:entityType];
    if (filter) {
        return [FilteredEnumerator enumeratorWithEnumerator:origEnum
                                                     filter:filter
                                                   selector:@selector(name)];
    } else {
        return origEnum;
    }
}

- (NSEnumerator *)enumeratorOfEntitiesTyped:(PajeEntityType *)entityType
                                inContainer:(PajeContainer *)container
                                   fromTime:(NSDate *)start
                                     toTime:(NSDate *)end
{
    NSEnumerator *subcontainerEnum;
    id subcontainer;
    MultiEnumerator *multiEnum;
    PajeContainerType *parentType;

    // if container should contain entityType directly, we don't filter!
    if ([[entityType containerType] isEqual:[container entityType]])
        return [super enumeratorOfEntitiesTyped:entityType
                                    inContainer:container
                                       fromTime:start
                                         toTime:end];

    // container does not contain entityType's directly.
    //Try containers in-between to see if they should be filtered.
    parentType = [entityType containerType];
    subcontainerEnum = [self enumeratorOfContainersTyped:parentType
                                             inContainer:container];
    multiEnum = [MultiEnumerator enumerator];
    while ((subcontainer = [subcontainerEnum nextObject]) != nil) {
        NSEnumerator *enumerator;
        enumerator = [super enumeratorOfEntitiesTyped:entityType
                                          inContainer:subcontainer
                                             fromTime:start
                                               toTime:end];
        if (enumerator != nil)
            [multiEnum addEnumerator:enumerator];
    }
    return multiEnum;
}

- (void)filterContainer:(PajeContainer *)container show:(BOOL)show
{
    id entityType = [container entityType];
    NSMutableSet *filter = [filters objectForKey:entityType];

    if (filter) {
        /*if (show)*/
         if ([filter containsObject:[container name]])
            [filter removeObject:[container name]];
        else
            [filter addObject:[container name]];
        [self registerDefaultFilter:filter forEntityType:entityType];
        [super dataChangedForEntityType:entityType];
    } else {
        [super filterContainer:container show:show];
    }
}

//
// methods to be a dragging destination, as delegate of matrix
//

- (void)highlightCell:(id)cell inMatrix:(NSMatrix *)m
{
    int r, c;
    static id highlightedCell; // the last cell that has been highlighted
    
    if (highlightedCell && ![highlightedCell isEqual:cell]) {
        [m getRow:&r column:&c ofCell:highlightedCell];
        [m highlightCell:NO atRow:r column:c];
        highlightedCell = nil;
    }
    
    if (cell && ![cell isEqual:highlightedCell]) {
        [m getRow:&r column:&c ofCell:cell];
        [m highlightCell:YES atRow:r column:c];
        highlightedCell = cell;
    }
}

- (unsigned int)matrix:(NSMatrix *)m draggingUpdated:(id <NSDraggingInfo>)sender
{
    NSPoint point = [m convertPoint:[sender draggingLocation] fromView:nil];
    NSButtonCell *cell = [m cellAtPoint:point];

    [self highlightCell:cell inMatrix:m];
    if (cell)
        return NSDragOperationGeneric;
    else
        return NSDragOperationNone;
}

#ifdef GNUSTEP
- (BOOL)matrix:(NSMatrix *)m prepareForDragOperation:(id <NSDraggingInfo>)sender
{
    return YES;
}
#endif

- (BOOL)matrix:(NSMatrix *)m performDragOperation:(id <NSDraggingInfo>)sender
{
    NSPoint point;
    NSButtonCell *cell;

    point = [m convertPoint:[sender draggingLocation] fromView:nil];
    cell = [m cellAtPoint:point];

    [self highlightCell:nil inMatrix:m];
    
    if (cell) {
        id entityType = [self selectedEntityType];
        [inputComponent setColor:[NSColor colorFromPasteboard:[sender draggingPasteboard]]
                         forName:[cell representedObject]
                    ofEntityType:entityType];
        if ([cell state] == 1)
            [super colorChangedForEntityType:entityType];
        [self synchronizeMatrix];
        return YES;
    }
    return NO;
}

- (void)matrix:(NSMatrix *)m draggingExited:(id <NSDraggingInfo>)sender
{
    [self highlightCell:nil inMatrix:m];
}

@end
