// DrawView+Finding
// --------------
// methods for finding entities in DrawView.

#include "DrawView.h"
#ifdef GNUSTEP
#include <math.h>
#include <values.h>
#endif

@implementation DrawView (Finding)
/*
 * Finding Entities
 * ------------------------------------------------------------------------
 */

BOOL line_hit(double px, double py,
              double x1, double y1, double x2, double y2)
// returns YES if (px,py) lies over the line from (x1,y1) to (x2,y2)
{
    double dx = px - x1;
    double dy = py - y1;
    double w = x2 - x1;
    double h = y2 - y1;

    if (fabs(w) > fabs(h)) {
        // horizontal line
        double ny;
        if (((dx * w) < 0)
            || (fabs(dx) > fabs(w)))
            return NO;
        ny = dx / w * h;
        if (fabs(ny - dy) <= 1)
            return YES;
    } else {
        // vertical line
        double nx;
        if (((dy * h) < 0)
            || (fabs(dy) > fabs(h)))
            return NO;
        nx = dy / h * w;
        if (fabs(nx - dx) <= 1)
            return YES;
    }
    return NO;
}

- (BOOL)isPoint:(NSPoint)p insideEntity:(PajeEntity *)entity
// return YES if the pixel at point "p" would be painted by drawing "entity".
{
    NSRect r;
    int hit;
#ifndef GNUSTEP
    pathfunction *path;
    STEntityDescriptor *entityDescriptor;
#endif

    // couldn't make PSinfill work with an arrow...
    switch ([filter drawingTypeForEntity:entity]) {
        case PajeLinkDrawingType:
            r = [self rectForEntity:entity];
            hit = line_hit(p.x, p.y, NSMinX(r), NSMinY(r), NSMaxX(r), NSMaxY(r));
            break;
        case PajeVariableDrawingType:
            r = [self rectForEntityType:[filter entityTypeForEntity:entity]
                            inContainer:[filter containerForEntity:entity]];
            r.origin.x   = TIMEtoX([filter startTimeForEntity:entity]);
            r.size.width = TIMEtoX([filter endTimeForEntity:entity])
                         - r.origin.x;
            hit = NSPointInRect(p, r);
            break;
        default:
            r = [self rectForEntity:entity];
#ifdef GNUSTEP
            hit = NSPointInRect(p, r);
#else
            // test insideness with x=y=0, because PostScript has a problem with
            // insideness testing in big coordinates (>~10000)
            entityDescriptor = [self descriptorForEntityType:
                                     [filter entityTypeForEntity:entity]];
            path = [entityDescriptor pathFunction];
            PSgsave();
            path(0, 0, NSWidth(r), NSHeight(r));
            PSinfill(p.x - NSMinX(r), p.y - NSMinY(r), &hit);
            PSgrestore();
#endif
    }

    return hit;
}


- (PajeEntity *)findEntityAtPoint:(NSPoint)point
// returns the entity that is drawn at "point", or nil if there isn't any.
// if there is more than one entity, returns the one that is on top (the
// last one to be drawn.
{
    return [self entityAtPoint:point
                    inInstance:[self rootInstance]
                  ofEntityType:[filter entityTypeForEntity:[self rootInstance]]];
}

- (PajeEntity *)entityAtPoint:(NSPoint)point
           inEntityDescriptor:(STEntityDescriptor *)edesc
                  inContainer:(PajeContainer *)container
{
    id findStartTime;
    id findEndTime;
    NSEnumerator *enumerator;
//    float closestDistance = MAXFLOAT;
    PajeEntity *closestEntity = nil;
    PajeEntity *entity;
    NSRect rect = [edesc rectInContainer:container];

    if ([edesc drawingType] != PajeLinkDrawingType) {
        NSRect r = rect;
        r.origin.x = point.x - 1;
        r.size.width = 2;
        if (!NSPointInRect(point, r))
            return nil;
    }

    findStartTime = XtoTIME(point.x + NSMinX(rect) - 1);
    findEndTime   = XtoTIME(point.x + NSMaxX(rect) + 1);
    enumerator = [filter enumeratorOfEntitiesTyped:[edesc entityType]
                                       inContainer:container
                                          fromTime:findStartTime
                                            toTime:findEndTime];
    while ((entity = [enumerator nextObject]) != nil) {
        if ([self isPoint:point insideEntity:entity]) {
            // find the entity whose center is closest to point
//            NSRect r = [self rectForEntity:entity];
//            float dx = NSMidX(r) - point.x;
//            float dy = NSMidY(r) - point.y;
//            float distance = dx * dx + dy * dy;
//            if (distance < closestDistance) {
//                closestDistance = distance;
//NSLog(@"closest:%@ -> %@", closestEntity, entity);
                closestEntity = entity;
break;
//            }
        }
    }
//    NSLog(@"search %@ in %@: %@", edesc, container, closestEntity);
    return closestEntity;
}

- (PajeEntity *)entityAtPoint:(NSPoint)point
                   inInstance:(id)instance
                 ofEntityType:(PajeEntityType *)entityType
{
    NSEnumerator *subtypeEnum;
    PajeEntityType *subtype;
    PajeEntity *entity;

    // search instances of each subhierarchy
    // (in reverse order, to find what's been drawn on top first)
    subtypeEnum = [[filter containedTypesForContainerType:entityType] reverseObjectEnumerator];
    while ((subtype = [subtypeEnum nextObject]) != nil) {
        entity = [self entityAtPoint:point
                         inContainer:instance
                        ofEntityType:subtype];
        if (entity != nil) {
            return entity;
        }
    }
    return nil;
}

- (PajeEntity *)entityAtPoint:(NSPoint)point
                  inContainer:(PajeContainer *)container
                 ofEntityType:(PajeEntityType *)entityType
{
    PajeEntity *entity;
    STEntityDescriptor *edesc = [self descriptorForEntityType:entityType];
    if (![edesc isContainer]) {
        entity = [self entityAtPoint:point
                  inEntityDescriptor:edesc
                         inContainer:container];
        if (entity) return entity;
    } else {
        if (container != nil || [edesc isPoint:point inContainer:container]) {
            NSEnumerator *ienum;
            id instance;
            // check all instances on this hierarchy
            ienum = [filter enumeratorOfContainersTyped:entityType
                                            inContainer:container];
            while ((instance = [ienum nextObject]) != nil) {
                if ([edesc isPoint:point inInstance:instance]) {
                    entity = [self entityAtPoint:point
                                      inInstance:instance
                                    ofEntityType:entityType];
                    if (entity != nil) {
                        return entity;
                    }
                }
            }
        }
    }
    return nil;
}


- (NSArray *)findEntitiesAtPoint:(NSPoint)point
// returns the entities that are drawn at "point", or nil if there isn't any.
{
    return [self entitiesAtPoint:point
                      inInstance:[self rootInstance]
                    ofEntityType:[filter entityTypeForEntity:[self rootInstance]]];
}

- (NSArray *)entitiesAtPoint:(NSPoint)point
          inEntityDescriptor:(STEntityDescriptor *)edesc
                 inContainer:(PajeContainer *)container
{
    NSDate *findStartTime;
    NSDate *findEndTime;
    NSEnumerator *enumerator;
    PajeEntity *entity;
    NSMutableArray *array;
    NSRect rect;

    rect = [edesc rectInContainer:container];

    if ([edesc drawingType] != PajeLinkDrawingType) {
        NSRect r = rect;
        r.origin.x = point.x - 1;
        r.size.width = 2;
        if (!NSPointInRect(point, r))
            return nil;
    }

    array = [NSMutableArray array];
    
    findStartTime = XtoTIME(point.x + NSMinX(rect) - 1);
    findEndTime   = XtoTIME(point.x + NSMaxX(rect) + 1);
    enumerator = [filter enumeratorOfEntitiesTyped:[edesc entityType]
                                       inContainer:container
                                          fromTime:findStartTime
                                            toTime:findEndTime];
    while ((entity = [enumerator nextObject]) != nil) {
        if ([self isPoint:point insideEntity:entity]) {
            [array addObject:entity];
        }
    }
    return array;
}

- (NSArray *)entitiesAtPoint:(NSPoint)point
                  inInstance:(id)instance
                ofEntityType:(PajeEntityType *)entityType
{
    NSEnumerator *subtypeEnum;
    PajeEntityType *subtype;
    NSArray *entities;
    NSMutableArray *array;

    array = [NSMutableArray array];
    // search instances of each subhierarchy
    // (in reverse order, to find what's been drawn on top first)
    subtypeEnum = [[filter containedTypesForContainerType:entityType] reverseObjectEnumerator];
    while ((subtype = [subtypeEnum nextObject]) != nil) {
        entities = [self entitiesAtPoint:point
                             inContainer:instance
                            ofEntityType:subtype];
        if (entities != nil) {
            [array addObjectsFromArray:entities];
        }
    }
    return array;
}

- (NSArray *)entitiesAtPoint:(NSPoint)point
                 inContainer:(PajeContainer *)container
                ofEntityType:(PajeEntityType *)entityType
{
    NSArray *entities;
    STEntityDescriptor *edesc;

    edesc = [self descriptorForEntityType:entityType];
    if (![edesc isContainer]) {
        return [self entitiesAtPoint:point
                  inEntityDescriptor:edesc
                         inContainer:container];
    } else {
        if (container == nil || [edesc isPoint:point inContainer:container]) {
            NSEnumerator *ienum;
            id instance;
            NSMutableArray *array;

            array = [NSMutableArray array];
            // check all instances on this hierarchy
            ienum = [filter enumeratorOfContainersTyped:entityType
                                            inContainer:container];
            while ((instance = [ienum nextObject]) != nil) {
                if ([edesc isPoint:point inInstance:instance]) {
                    entities = [self entitiesAtPoint:point
                                          inInstance:instance
                                        ofEntityType:entityType];
                    if (entities != nil) {
                        [array addObjectsFromArray: entities];
                    }
                }
            }
            return array;
        }
    }
    return nil;
}

@end
